import traceback

class print_exc(object):
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == KeyboardInterrupt:
            return
        if exc_type:
            print ''.join(traceback.format_exception(exc_type, exc_val, exc_tb))
        return True

print_exc = print_exc()


import importlib
import sys
import urllib

# Args
assert len(sys.argv) == 4, 'Usage: load_task.py host path entry'
host = sys.argv[1]
path = sys.argv[2]
entry = sys.argv[3]

# Import Functions
sys.path.insert(0, path)
module, _, fun = entry.rpartition('.')
module = importlib.import_module(module)
fun = getattr(module, fun)
assert hasattr(fun, 'paxd_queue'), 'This function is not a paxd task'
queue = fun.paxd_queue

url = 'http://{host}:22100/load?path={path}&entry={entry}&queue={queue}'.format(**vars())

print urllib.urlopen(url).read()


import sys
import pickle
from redis import Redis
from paxd.app.http import Request
from paxd.process import print_exc
from multiprocessing import Pipe, Process
from threading import Thread
from Queue import Empty
import uuid
from signal import signal, SIGINT
from paxd.app.msg import PRequest #, BaseResponse

def start_webui(broker_host, broker_port, name):
    WebUI(name, broker_host, broker_port).start()
    
class WebUI(object):
    def __init__(self, name, broker_host, broker_port):
        self.name = name
        self.broker_host = broker_host
        self.broker_port = broker_port
    
    def start(self):
        http_server = Thread(target=start_server, args=(self.broker_host, self.broker_port, self.name))
        http_server.daemon = True
        http_server.start()
        # signal(SIGINT, sigint_handler)
        while True:
            try:
                http_server.join(0.5)
            except KeyboardInterrupt:
                print 'WebUI listener process exiting\n'
                sys.exit(-1)

from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

shut_down = None

def start_server(broker_host, broker_port, name):
    redis = Redis(broker_host, port=broker_port)
    print 'SENDING REQUESTS TO', name
    class WebUIHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            if self.bad_path():
                return self.do_404()
            
            request = Request(self.client_address, 
                                self.command, 
                                self.path, 
                                self.request_version, 
                                [])
            
            paxd_req = PRequest(redis, name, args=[request], timeout=30)
            try:
                response = paxd_req.send().get()
            except Exception, e:
                print e
                print e.traceback
                raise
            if not response:
                self.send_response(500)
                self.end_headers()
                self.wfile.write('TIMED OUT')
                return
            self.send_response(response.code)
            if response.content_type:
                self.send_header('Content-Type', response.content_type)
            if response.location:
                self.send_header('Location', response.location)
            self.end_headers()
            self.wfile.write(response.body)
        
        def bad_path(self):
            return 'favicon' in self.path
        
        def do_404(self):
            self.send_response(404)
            self.end_headers()
    
    server_class = HTTPServer
    handler_class = WebUIHandler
    server_address = ('', 22100)
    print 'listening on', server_address
    httpd = server_class(server_address, handler_class)
    httpd.serve_forever()

def sigint_handler(num, frame):
    import sys
    print >>sys.stderr, 'EXITING'
    sys.exit(0)


import urllib
import os
import random
import string
import tarfile
import shutil

def random_name():
    return ''.join([random.choice(string.letters) for i in range(0, 32)])

def download_temp(url):
    root = '/tmp/paxd-%s' % random_name()
    os.makedirs(root)
    archive = os.path.join(root, random_name() + '-archive.tar.gz')
    if url.startswith('file://'):
        shutil.copyfile(url[7:], archive)
    else:
        urllib.urlretrieve(url, archive)
    tar = tarfile.open(archive, mode='r:gz')
    tar.extractall(root)
    tar.close()
    os.unlink(archive)
    return root

def join(root, file):
    if file and file[0] == '/':
        file = file[1:]
    file = os.path.abspath(os.path.join(root, file))
    assert file.startswith(root), 'file must be within root directory'
    return file




import urlparse

class Request(object):
    def __init__(self, client_address, command, path, request_version, headers):
        self.client_address = client_address
        self.command = command
        self.path = path
        self.request_version = request_version
        self.headers = headers
        self.GET = self.parse_GET(self.path)
    
    @classmethod
    def parse_GET(cls, path):
        parts = urlparse.urlparse(path)
        ret = {}
        for k, l in urlparse.parse_qs(parts.query, keep_blank_values=True).iteritems():
            ret[k] = l[0]
        return ret
    

class Response(object):
    def __init__(self, code, body, content_type=None, location=None):
        self.code = code
        self.body = body
        self.content_type = content_type
        self.location = location




import time
from datetime import datetime
from multiprocessing import Process, Pipe, Queue
from Queue import Empty
from signal import signal, SIGINT, SIGKILL, SIGTERM
from threading import RLock, Thread
import os
import sys
import pickle
import json
from redis import Redis

import threading
from paxd.app.application import Application
from paxd.process import print_exc
from paxd.server.webserver import start_webui
from paxd.server import fsys
from paxd.app.msg import BaseRequest, BaseResponse
from paxd import PROJECT_PATH, VERSION

class Controller(object):
    command_queue = '_paxd.controller'
    webui_queue = '_paxd.web_ui'
    def __init__(self, instance_id, broker_host, broker_port):
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.instance_id = instance_id
        self.lock = RLock()
        self.processes = []
        self.threads = []
        self.applications = {}
        self.initialized = False
        self.application_history = []
    
    def run(self):
        try:
            print 'STARTING UP CONTROLLER.  pid: ', os.getpid()
            # Internal Applications
            self.start_process(start_webui, '%s~%s' % (self.webui_queue, VERSION))
            
            # Start "apps"
            self.configure()
            
            # useful for debugigng, prints out stack traces every 4 seconds
            # for things which aren't sleeping or blocking on locks/network
            # self.introspector = Introspector()
            # self.introspector.daemon = True
            # self.introspector.start()
            
            # Wait For Commands
            self.handle_commands()
        except Exception, e:
            print 'TERMINATING', e
            import sys
            for process in self.processes:
                process.terminate()
            raise
    
    def configure(self):
        if self.initialized:
            return
        print self.broker_host, self.broker_port
        redis = Redis(self.broker_host, port=self.broker_port)
        config = redis.hget('_paxd.controllers.config', self.instance_id)
        self.load_web_ui().start()
        if not config:
            return
        
        config = json.loads(config)
        for app in config['applications']:
            if app['name'] == 'Pax Daemonica':
                continue
            print 'LOADING', app['name']
            
            app = Application.from_conf(self.broker_host, self.broker_port, app['data_root'], app, self.instance_id)
            assert app.id not in self.applications
            self.applications[app.id] = app
        
        for time, app in config.get('application_history',[]):
            print 'CREATING HISTORIC ENTRY FOR', app['name']
            app = Application.from_conf(self.broker_host, self.broker_port, app['data_root'], app, self.instance_id)
            self.application_history.append((time, app))

        
        self.initialized = True
    
    def handle_commands(self):
        thread = Thread(target=command_thread, args=(self,))
        thread.daemon = True
        thread.start()
        self.threads.append(thread)
        while True:
            try:
                thread.join(60)
            except KeyboardInterrupt:
                print 'CONTROLLER EXITING'
                sys.exit(0)
    
    def load_web_ui(self):
        app = Application.from_conf(self.broker_host, self.broker_port, PROJECT_PATH, WEBUI_CONF, self.instance_id, system=True)
        if app.id not in self.applications:
            self.applications[app.id] = app
        self.applications[app.id].unpause()
        return self.applications[app.id]
        
    def load_app(self, url, config_file='/paxd.conf', group=None):
        path = self.setup_workspace(url)
        data = self.load_config(path, config_file)
        app = Application.from_conf(self.broker_host, self.broker_port, path, data, self.instance_id, group=group)
        if app.id in self.applications:
            raise Exception('App ID already exists')
        self.applications[app.id] = app
        return app
    
    def update_app(self, name, group, version, url, config_file):
        id = Application.make_id(name=name, group=group, version=version)
        assert id in self.applications
        path = self.setup_workspace(url)
        data = self.load_config(path, config_file)
        app = self.applications[id]
        app.update(path, data, self.instance_id)
        return app
    
    def start_app(self, name, group='default', version=''):
        id = Application.make_id(name=name, group=group, version=version)
        assert id in self.applications
        self.applications[id].start()
    
    def start_process(self, fun, name):
        process = Process(target=fun, args=(self.broker_host, self.broker_port, name))
        process.start()
        print 'STARTING INTERNAL PROCESS', name, process.pid
        self.processes.append(process)
    
    def setup_workspace(self, url):
        return fsys.download_temp(url)
    
    def load_config(self, path, config_file):
        conf = fsys.join(path, config_file)
        with open(conf) as f:
            data = f.read()
            return json.loads(data)


def command_thread(controller):
    redis = Redis(controller.broker_host, port=controller.broker_port)
    while True:
        with print_exc:
            _, value = redis.blpop(controller.command_queue)
            request = BaseRequest.from_serialized(value)
            
            resp_data = handle_command(controller, redis, request.deserialize(request.args))
            
            response = request.response_class(request.response_queue, 'SUCCESS', resp_data)
            pipe = redis.pipeline(transaction=True)
            q = request.response_queue
            ser = response.serialize_resp()
            pipe.lpush(q, ser)
            pipe.expire(request.response_queue, request.timeout)
            # pipe.lrem(self.active_queue, 1, request.raw)
            pipe.execute()
            save_settings(redis, controller)

def save_settings(redis, controller):
    pipe = redis.pipeline(transaction=True)
    apps = []
    with controller.lock:
        for id, app in controller.applications.iteritems():
            apps.append(app.json_info())
    data = {
        'applications' : apps,
        'application_history' : [(t, a.json_info()) for t, a in controller.application_history]
    }
    # import pprint
    # pprint.pprint(data)
    redis.hset('_paxd.controllers.config', controller.instance_id, json.dumps(data))

def handle_command(controller, redis, rdata):
    try:
        command = rdata['command']
        if command == 'APPLIST':
            return handle_list(controller, redis, rdata)
        elif command == 'LOAD_APP':
            return handle_load_app(controller, redis, rdata)
        elif command == 'UPDATE_APP':
            return handle_update_app(controller, redis, rdata)
        elif command == 'LOAD_AUTO':
            return handle_load_auto(controller, redis, rdata)
        elif command == 'UNLOAD_APP':
            return handle_unload(controller, redis, rdata)
        elif command == 'PAUSE' or command == 'UNPAUSE':
            return handle_pause_unpause(controller, redis, rdata)
        elif command == 'REQUEUE_ALL_FAILED':
            return handle_requeue_all_failed(controller, redis, rdata)
        elif command == 'REMOVE_ALL_FAILED':
            return handle_remove_all_failed(controller, redis, rdata)
        elif command == 'TRACE_COMMAND':
            return handle_trace_command(controller, redis, rdata)
        elif command == 'TRACE_APP':
            return handle_trace_app(controller, redis, rdata)
        elif command == 'GET_FAILED':
            return handle_get_failed(controller, redis, rdata)
        elif command == 'REQUEUE_FAILED':
            return handle_requeue_failed(controller, redis, rdata)
        elif command == 'REMOVE_FAILED':
            return handle_remove_failed(controller, redis, rdata)
        return 'BAD COMMAND'
    except Exception, e:
        import traceback
        print e
        traceback.print_exc()
        return { 'status' : 'ERROR', 'exception' : e }

def handle_list(controller, redis, rdata):
    ret = []
    with controller.lock:
        for id, app in controller.applications.iteritems():
            json_info = app.json_info(detailed=True)
            ret.append(json_info)
        
        historic = []
        for time, app in controller.application_history:
            info = app.json_info()
            info['terminated'] = time
            historic.append(info)
    
    return {'applications' : ret, 'history' : historic }

def handle_trace_app(controller, redis, rdata):
    app = get_app(controller, rdata)
    if app is None:
        return NO_APP_MSG
    return { 'status' : 'SUCCESS', 'traces' : app.trace, 'app' : app.json_info() }

def handle_trace_command(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    return { 'status' : 'SUCCESS', 'traces' : command.trace, 'app' : app.json_info() }

def handle_requeue_failed(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    res = command.requeue_failed(rdata['fail_id'])
    return { 'status' : 'SUCCESS' }

def handle_remove_failed(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    res = command.remove_failed(rdata['fail_id'])
    return { 'status' : 'SUCCESS' }

def handle_get_failed(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    return { 'status' : 'SUCCESS', 'failed' : command.failed_items, 'app' : app.json_info(), 'command' : command.json_info() }

def handle_load_app(controller, redis, rdata):
    app = controller.load_app(rdata['url'], config_file=rdata['config_file'], group=rdata.get('group'))
    return { 'status' : 'SUCCESS', 'apps' : [app.json_info()] }

def handle_update_app(controller, redis, rdata):
    app = controller.update_app(name=rdata['name'],
                                group=rdata['group'],
                                version=rdata['version'],
                                url=rdata['url'],
                                config_file=rdata['config_file'])
    return { 'status' : 'SUCCESS', 'apps' : [app.json_info()] }

def handle_unload(controller, redis, rdata):
    with controller.lock:
        app = get_app(controller, rdata)
        if app is None:
            return { 'status' : 'FAILURE', 'reason' : 'NO SUCH APP ID' }
        app.quit()
        controller.application_history.append((datetime.utcnow().ctime(), app))
        del controller.applications[app.id]
        return { 'status' : 'SUCCESS', 'app' : app.json_info() }

def handle_requeue_all_failed(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    command.requeue_all()
    return { 'status' : 'SUCCESS', 'app' : app.json_info() }
    
def handle_remove_all_failed(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    command.remove_all_failed()
    return { 'status' : 'SUCCESS', 'app' : app.json_info() }

def handle_pause_unpause(controller, redis, rdata):
    app = get_app(controller, rdata)
    command = get_command(controller, app, rdata)
    if app is None or command is None:
        return NO_APP_MSG
    if rdata['command'] == 'UNPAUSE':
        command.unpause()
    else:
        command.pause()
    return { 'status' : 'SUCCESS', 'app' : app.json_info() }

NO_APP_MSG = { 'status' : 'FAILURE', 'reason' : 'NO SUCH APP ID' }

def get_command(controller, app, rdata):
    if rdata['queue'] in app:
        return app[rdata['queue']]
    return None

def get_app(controller, rdata):
    id = Application.make_id(name=rdata['name'], group=rdata['group'], version=rdata['version'])
    with controller.lock:
        if id in controller.applications:
            return controller.applications[id]
    return None
    

WEBUI_CONF = {
    "name" : "Pax Daemonica", 
    "group" : "default", 
    "python_paths" : ["/"], 
    "required_packages" : [] ,
    "commands" : {
        "_paxd.web_ui" : {
            "entry" : "paxd.webuiapp.webui:webui_app"
        },
    },
    "environment" : {},
    "version" : VERSION
}


class Introspector(threading.Thread):
    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        
    def run(self):
        print 'start'
        while True:
            time.sleep(3)
            self.introspect()
            # break
    
    def introspect(self):
        print 'introspect'
        import traceback
        current_thread = threading.current_thread().ident
        try:
            import os
            stack = None
            res = []
            print '*' * 160 * 2
            for id, stack in sys._current_frames().items():
                data = ''.join(traceback.format_stack(stack))
                if 'task = get()' in data:
                    continue
                if 'waiter.acquire()' in data:
                    continue
                if 'time.sleep(' in data:
                    continue
                if 'data = self._sock.recv(self._rbufsize)' in data:
                    continue
                if 'traceback.format_stack(stack)' in data:
                    continue
                if '_sleep(delay' in data:
                    continue
                
                print '-' * 80
                print id
                print data
        finally:
            del stack



import time
import uuid
import pickle
import json
from redis import Redis
from app import exclaimer


if __name__ == '__main__':
    rqs = []
    for i in range(0, 2000):
        promise = exclaimer.delay(str(i), paxd_timeout=4)
        rqs.append(promise)
    
    for pr in rqs:
        print pr.get()
        # try:
        #     rqs.append(request.send())
        # except Exception, e:
        #     print e.type, e.traceback
        # time.sleep(1)
        # str(i),
        #     'response_queue' : rq,
        # }))



import os
import uuid
import traceback
import json
import pickle
from paxd.app.http import Response
from paxd.app.msg import PRequest
from redis import Redis
from paxd.webuiapp import html

################################
#  Web UI App/Routing
################################

def webui_app(request):
    broker_host = os.environ['BROKER_HOST']
    broker_port = int(os.environ.get('BROKER_PORT', 6379))
    try:
        response = handle_request(request, broker_host, broker_port)
    except:
        response = Response(500, 'Server Error %s' % traceback.format_exc())
    return response


def route_mapping():
    return (
        ('/jquery.js', jquery),
        ('/load', load_app),
        ('/update', update_app),
        ('/unload', unload_app),
        ('/failed/remove', remove_failed),
        ('/failed/remove-all', remove_all_failed),
        ('/failed/requeue', requeue_failed),
        ('/failed/requeue-all', requeue_all_failed),
        ('/failed', get_failed),
        ('/pause', pause_app),
        ('/unpause', unpause_app),
        ('/trace_command', trace_command),
        ('/trace_app', trace_app),
        ('/', get_apps),
    )

def route(request):
    for path, fun in route_mapping():
        if request.path.partition('?')[0] == path:
            return fun
    return None

def handle_request(request, broker_host, broker_port):
    fun = route(request)
    if fun is None:
        return Response(404, 'Not Found')
    try:
        redis = Redis(broker_host, port=broker_port)
        return Response(200, fun(request, redis), content_type='text/html')
    except Forward, f:
        return Response(301, '', location=f.location)
    except NoResponse:
        return Response(500, 'No Response from controller')

################################
#  Web UI Actions
################################

controller_queue = '_paxd.controller'
def get_apps(request, redis):
    req = PRequest(redis, controller_queue, {
            'command' : 'APPLIST'
        }, timeout=2)
    promise = req.send()
    resp = promise.get()
    if not resp:
        return 'NO RESPONSE'
    
    return html.render('applications.html', resp)
    # return html.header() + html.applications(resp) + html.footer()

def jquery(request, redis):
    import os
    from paxd import PROJECT_PATH
    from mako.lookup import TemplateLookup
    
    path = os.path.join(PROJECT_PATH, 'paxd/webuiapp/templates/jquery-1.5.min.js')
    with open(path) as f:
        return f.read()

def load_app(request, redis):
    group = request.GET.get('group')
    
    conf = request.GET.get('config_file')
    if not conf:
        conf = '/paxd.conf'
    
    promise = PRequest(redis, controller_queue, {
        'command':'LOAD_APP',
        'url' : request.GET['url'],
        'group' : group, 
        'config_file' : conf, 
    }, timeout=60)

    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    raise Forward('/')

def update_app(request, redis):
    group = request.GET.get('group')
    
    conf = request.GET.get('config_file')
    if not conf:
        conf = '/paxd.conf'
    
    promise = PRequest(redis, controller_queue, {
        'command':'UPDATE_APP',
        'name' : request.GET['name'], 
        'version' : request.GET['version'], 
        'url' : request.GET['url'],
        'group' : group, 
        'config_file' : conf, 
    }, timeout=60)
    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    raise Forward('/')


def requeue_failed(request, redis):
    promise = PRequest(redis, controller_queue, {
        'name' : request.GET['name'], 
        'group' : request.GET['group'], 
        'version' : request.GET['version'],
        'queue' : request.GET['queue'], 
        'command' : 'REQUEUE_FAILED',
        'fail_id' : request.GET['fail_id'], 
    }, timeout=4)
    res = promise.send().get()
    if res is None:
        raise NoResponse()
    return json.dumps(res)

def remove_failed(request, redis):
    promise = PRequest(redis, controller_queue, {
        'name' : request.GET['name'], 
        'group' : request.GET['group'], 
        'version' : request.GET['version'], 
        'queue' : request.GET['queue'], 
        'command' : 'REMOVE_FAILED',
        'fail_id' : request.GET['fail_id'], 
    }, timeout=4)
    res = promise.send().get()
    if res is None:
        raise NoResponse()
    return json.dumps(res)

def get_failed(request, redis):
    resp = command_action(redis, request, 'GET_FAILED')
    res = html.render('failed.html', resp)
    return res

def unload_app(request, redis):
    resp = app_action(redis, request, 'UNLOAD_APP')
    return json.dumps(resp)

def pause_app(request, redis):
    resp = command_action(redis, request, 'PAUSE')
    return json.dumps(resp)

def unpause_app(request, redis):
    resp = command_action(redis, request, 'UNPAUSE')
    return json.dumps(resp)

def remove_all_failed(request, redis):
    resp = command_action(redis, request, 'REMOVE_ALL_FAILED')
    return json.dumps(resp)

def requeue_all_failed(request, redis):
    resp = command_action(redis, request, 'REQUEUE_ALL_FAILED')
    return json.dumps(resp)


def trace_app(request, redis):
    resp = app_action(redis, request, 'TRACE_APP')
    return html.render('traces.html', resp)

def trace_command(request, redis):
    resp = command_action(redis, request, 'TRACE_COMMAND')
    return html.render('traces.html', resp)

def app_action(redis, request, command_name):
    promise = PRequest(redis, controller_queue, {
        'command' : command_name,
        'name' : request.GET['name'], 
        'group' : request.GET['group'], 
        'version' : request.GET['version'], 
        # 'queue' : request.GET['queue'], 
    }, timeout=4)
    res = promise.send().get()
    if res is None:
        raise NoResponse()
    return res

def command_action(redis, request, command_name):
    promise = PRequest(redis, controller_queue, {
        'command' : command_name,
        'name' : request.GET['name'], 
        'group' : request.GET['group'], 
        'version' : request.GET['version'], 
        'queue' : request.GET['queue'], 
    }, timeout=4)
    res = promise.send().get()
    if res is None:
        raise NoResponse()
    return res
    

# def id_app(request, command_name):
#     redis = Redis('localhost')
#     promise = PRequest(redis, controller_queue, {
#         'command' : command_name,
#         'id' : request.GET['id'], 
#     }, timeout=4)
#     res = promise.send().get()
#     if res is None:
#         raise NoResponse()
#     return res

class NoResponse(Exception):
    pass

class Forward(Exception):
    def __init__(self, location):
        self.location = location


import os
import json
from redis import Redis
from functools import wraps
from paxd.app.msg import PRequest

def queue_name(queue, version):
    return '%s~%s' % (queue, version)

def make_task_file(path, broker_host=None, broker_port=None):
    with open(path) as f:
        data = json.loads(f.read())
    return make_task_conf(data, broker_host=broker_host, broker_port=broker_port)

def make_task_conf(conf, broker_host=None, broker_port=None):
    if broker_host is None:
        broker_host = os.environ['BROKER_HOST']
    if broker_port is None:
        broker_port = int(os.environ.get('BROKER_PORT', 6379))
    redis = Redis(broker_host, port=broker_port)
    class task(object):
        def __init__(self, queue):
            self._queue = queue
            self.queue = queue_name(queue, conf.get('version',''))
            self.conf = conf
        
        def __call__(self, fun):
            @wraps(fun)
            def delay(*args, **kwargs):
                if 'paxd_timeout' in kwargs:
                    paxd_timeout = kwargs['paxd_timeout']
                    del kwargs['paxd_timeout']
                else:
                    paxd_timeout = None
                return PRequest(redis, self.queue, args=args, kwds=kwargs, timeout=paxd_timeout).send()
            @wraps(fun)
            def run_offline(*args, **kwargs):
                if 'paxd_timeout' in kwargs:
                    paxd_timeout = kwargs['paxd_timeout']
                    del kwargs['paxd_timeout']
                else:
                    paxd_timeout = None
                return PRequest(redis, self.queue, args=args, kwds=kwargs, timeout=paxd_timeout, offline=True).send()
            fun.delay = delay
            fun.run_offline = run_offline
            fun.paxd_task = True
            fun.paxd_queue = self.queue
            return fun
    return task


import traceback
import sys
import time
from threading import RLock, Thread
import uuid
from redis import Redis
from multiprocessing import Process, Pipe, cpu_count
from paxd.process import print_exc
from paxd.app.msg import BaseRequest, BaseResponse
from paxd.server import fsys

from smartpool.pool import Pool
from paxd.app.command import Command, init_process

class Application(object):
    DEFAULT_WORKERS = 2
    
    def __init__(self, broker_host, broker_port, name, data_root, env, packages, 
                    commands, num_workers, version, 
                    python_paths, group, system, state='stopped'):
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.lock = RLock()
        self.data_root = data_root
        self.env = env
        self.group = group
        self.id = self.make_id(name=name, group=group, version=version)
        self.name = name
        self.num_workers = num_workers
        self.packages = packages
        self.python_paths = python_paths
        self.state = state
        self.system = system
        self.version = version
        self.exit = False
        
        # Commands
        self.commands = commands
        self.old_commands = []
        
        # Shared Pool
        self.pool = Pool(processes=self.num_workers, initializer=init_process, initargs=[self.abs_python_paths(), self.env])
        self.old_pools = []
        self.shared_pool = SharedPool(self, self.pool, self.old_pools)
        self.shared_pool.daemon = True
        for q, c in self.commands.items():
            c.set_shared_size(self.num_workers)
        
        # Cleanup thread
        self.cleaner = CleanUp(self)
        self.cleaner.daemon = True
        self.cleaner.start()
        
        # start if "started" state passed in
        if self.state == 'started':
            self.unpause()
    
    def quit(self):
        self.pause()
        self.resize(0)
        self.exit = True
        for q, c in self.commands.iteritems():
            c.quit()
    
    def pause(self):
        for queue, command in self.commands.iteritems():
            command.pause()
        self.state = 'stopped'
    
    def unpause(self):
        self.start()
        for queue, command in self.commands.iteritems():
            command.unpause()
        self.state = 'started'
    
    def resize(self, size):
        self.pool.resize(size)
        self.num_workers = size
        for q, c in self.commands.items():
            c.set_shared_size(self.num_workers)
    
    @property
    def trace(self):
        return self.pool.introspect()
    
    def __getitem__(self, queue):
        return self.commands[queue]
    
    def __contains__(self, queue):
        return queue in self.commands
    
    def abs_python_paths(self):
        return [fsys.join(self.data_root, p) for p in self.python_paths]
    
    def start(self):
        for queue, command in self.commands.iteritems():
            command.start()
        if not self.shared_pool.is_alive():
            self.shared_pool.start()
        if not self.cleaner.is_alive():
            self.cleaner.start()
    
    def stop(self):
        if self.system:
            return
    
    @staticmethod
    def make_id(**kwargs):
        return (kwargs['name'], kwargs['group'], kwargs['version'])
    
    @classmethod
    def from_conf(cls, broker_host, broker_port, path, conf, instance_id, system=False, group=None):
        app_args = {
            'name' : conf['name'],
            'group' : conf.get('group', 'default'),
            'data_root' : path,
            'env' : conf.get('environment', {}),
            'packages' : conf.get('required_packages',[]),
            'python_paths' : conf.get('python_paths',[]),
            'num_workers' : 2, # TODO, configurable
            'version' : conf.get('version',''),
            'system' : system,
        }
        
        if group != None:
            app_args['group'] = group
        
        if 'state' in conf:
            app_args['state'] = conf['state']
        
        app_args['commands'] = {}
        for queue, c in conf.get('commands', {}).iteritems():
            com = Command.from_conf(broker_host, broker_port, queue, c, path, app_args['python_paths'], app_args['env'], instance_id, 
                    shared_size=app_args['num_workers'], version=app_args['version'])
            app_args['commands'][queue] = com
        
        return Application(broker_host, broker_port, **app_args)
    
    def update(self, path, conf, instance_id):
        '''
            create new pool with new parameters
        '''
        assert self.version == conf.get('version',''), (self.version, conf.get('version',''))
        assert self.name == conf.get('name'), (self.name, conf.get('name'))
        
        final_state = self.json_info
        
        self.data_root = path
        self.env = conf.get('environment', {})
        self.packages = conf.get('required_packages',[])
        self.python_paths = conf.get('python_paths', [])
        
        # commands
        old_commands = self.commands
        
        self.commands = {}
        for queue, c in conf.get('commands', {}).iteritems():
            com = Command.from_conf(broker_host, broker_port, queue, c, self.data_root, self.python_paths, self.env, instance_id, 
                    shared_size=self.DEFAULT_WORKERS, version=self.version)
            self.commands[queue] = com
        for q, c in old_commands.iteritems():
            if q not in self.commands:
                continue
            if not c.stopped:
                self.commands[q].unpause()
        self.cleaner.add_commands(old_commands)
        
        if self.state == 'started':
            self.unpause()
        
        # Shared Pool
        with self.lock:
            self.cleaner.add_pool(self.pool)
            self.pool = Pool(processes=self.num_workers, initializer=init_process, initargs=[self.abs_python_paths(), self.env])
        
        for q, c in self.commands.items():
            c.set_shared_size(self.num_workers)
        
        # start if "started" state passed in
        if self.state == 'started':
            self.unpause()

        
    
    def json_info(self, detailed=False):
        data = {
            'data_root' : self.data_root,
            'environment' : self.env,
            'group' : self.group,
            # 'id' : self.id,
            'name' : self.name,
            'num_workers' : self.num_workers,
            'packages' : self.packages,
            'python_paths' : self.python_paths,
            'state' : self.state,
            'system' : self.system,
            'version' : self.version,
            'commands' : {},
        }
        for q, c in self.commands.iteritems():
            data['commands'][q] = c.json_info(detailed=detailed)
        return data
    
    def __repr__(self):
        return 'Application(name="%s", group="%s", version="%s")' % (self.name, self.group, self.version)

class CleanUp(Thread):
    def __init__(self, app):
        Thread.__init__(self)
        self.app = app
        self.lock = RLock()
        self.old_pools = []
        self.old_commands = []
    
    def add_pool(self, pool):
        pool.close()
        with self.lock:
            self.old_pools.append(pool)
    
    def add_commands(self, commands):
        for q, c in commands.iteritems():
            c.quit()
        with self.lock:
            self.old_commands.append(commands)
    
    def run(self):
        while True:
            with print_exc:
                if self.app.exit:
                    break
                self.event_loop()
    
    def event_loop(self):
        with self.lock:
            sleep = len(self.old_pools) == 0 and len(self.old_commands) == 0
        if sleep:
            time.sleep(2)
            return
        
        with self.lock:
            new_pools = []
            for p in self.old_pools:
                for i in p.introspect():
                    if i['state'] != 'EXITED':
                        new_pools.append(p)
                        break
            self.old_pools = new_pools
            new_commands = []
            # all old sets of commands
            for c_map in self.old_commands:
                finished = []
                # a single set of commands
                for q, c in c_map.iteritems():
                    for i in c.trace:
                        if i['state'] != 'EXITED':
                            break
                    else:
                        finished.append(q)
                for q in finished:
                    del c_map[q]
                if len(c_map) > 0:
                    new_commands.append(c_map)
            self.old_commands = new_commands

class SharedPool(Thread):
    def __init__(self, app, pool, old_pools, *args, **kwargs):
        Thread.__init__(self, *args, **kwargs)
        self.app = app
        self.pool = app.pool
        self.old_pools = old_pools
        self.pending = []
        self.lock = RLock()
    
    def full(self):
        with self.lock:
            self.pending = [x for x in self.pending if not x['async'].ready()]
            return len(self.pending) > self.app.num_workers
    
    def run(self):
        while True:
            with print_exc:
                if self.app.exit:
                    break
                self.event_loop()
    
    def event_loop(self):
        if self.full():
            time.sleep(0.1)
            return
        for queue, command in self.app.commands.items():
            if command.stopped:
                continue
            if command.runner.shared_ready():
                with self.lock:
                    res = command.runner.run_fun(pool=self.pool, pool_type='shared')
                    self.pending.append(res)
                return
        else:
            time.sleep(0.1)


import os
from paxd import PROJECT_PATH
import mako
from mako.lookup import TemplateLookup

def render(path, context):
    lookup = TemplateLookup(directories=[os.path.join(PROJECT_PATH, 'paxd/webuiapp/templates')])
    template = lookup.get_template(path)
    try:
        return template.render_unicode(**context)
    except:
        return mako.exceptions.html_error_template().render()
    



import time
import os
import os.path
# from appserver import Application
from paxd.server.controller import Controller

def main(id):
    assert 'BROKER_HOST' in os.environ, 'BROKER_HOST is required as an environment variable'
    broker_host = os.environ['BROKER_HOST']
    broker_port = int(os.environ.get('BROKER_PORT', 6379))
    c = Controller(id, broker_host, broker_port=broker_port)
    c.run()


if __name__ == '__main__':
    import sys
    main(sys.argv[1])
    


import time
import sys
sys.path.insert(0, '.')
import paxd.client
from paxd.client import make_task_file
import os.path 

task = make_task_file('testapp/paxd.conf')

@task(queue='testapp.app:exclaimer')
def exclaimer(thing):
    time.sleep(0.2)
    return str(thing) + '!!!!!!!'

@task(queue='testapp.app:error_test')
def error_test():
    raise Exception('BAD BAD BAD')



#!/usr/bin/env python

from distutils.core import setup

VERSION = '0.5'
DESCRIPTION = 'Python Application/Worker Server'

setup(
    name='PaxDaemonica',
    version=VERSION,
    description=DESCRIPTION,
    author='Jeffrey Jenkins',
    license='MIT',
    zip_safe=False,
    author_email='jeff@qcircles.net',
    url='http://github.com/jeffjenkins/PaxDaemonica',
    packages=['paxd', 'paxd.server', 'paxd.monit', 'paxd.app', 'paxd.webuiapp', 'paxd.webuiapp.templates'],
    package_data={
        "" : ["*.js", "*.css", "*.sass", "*.html"],
    },
    install_requires=['redis', 'smartpool', 'mako'],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # 'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)


import traceback
import importlib
from paxd.app.msg import BaseRequest, BaseResponse
from paxd.process import print_exc
from paxd.server import fsys
import time
from smartpool.pool import Pool
from Queue import Queue, Empty
from threading import RLock, Thread
from redis import Redis
from multiprocessing import Process, Pipe, cpu_count

WORKER_FUNCTIONS = {}
def worker_fun(entry, args, kwds, format):
    import json
    import cPickle as pickle
    if format == 'pickle':
        deserialize = pickle.loads
    else:
        deserialize = json.loads
    try:
        if entry not in WORKER_FUNCTIONS:
            module, _, fun = entry.rpartition(':')
            module = importlib.import_module(module)
            wfun = module
            for name in fun.split('.'):
                wfun = getattr(wfun, name)
            WORKER_FUNCTIONS[entry] = wfun
        args = deserialize(args)
        kwds = deserialize(kwds)
        return WORKER_FUNCTIONS[entry](*args, **kwds)
    except KeyboardInterrupt:
        time.sleep(0.3)
        raise
    except Exception, e:
        e.traceback = traceback.format_exc()
        raise e

def init_process(paths, env):
    import sys
    import os
    for p in reversed(paths):
        sys.path.insert(0, p)
    for k, v in env.iteritems():
        os.environ[k] = v

class Command(object):
    def __init__(self, broker_host, broker_port, queue, entry, base_path, python_paths, environment, instance_id, shared_size, state='stopped', num_processes=max(cpu_count(), 2), version=''):
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.entry = entry
        self.queue = queue
        self.base_path = base_path
        self.python_paths = python_paths
        self.environment = environment
        self.instance_id = instance_id
        self.redis = Redis(broker_host, port=broker_port)
        self.lock = RLock()
        
        # computed fields
        self.abs_python_paths = [fsys.join(base_path, p) for p in python_paths]
        
        self.queue_name = '%s~%s' % (queue, version)
        self.active_queue = '%s:active' % self.queue_name
        self.failed_queue = '%s:failed' % self.queue_name
        self.success_count = '%s:success' % self.queue_name
        self.pending_queue = '%s:pending' % self.queue_name
        
        # state
        self.exit = False
        self.state = state
        self.num_processes = num_processes
        self.pool = Pool(processes=self.num_processes, initializer=init_process, initargs=[self.abs_python_paths, environment])
        self.initialized = False
        self.shared_size = shared_size
        
        # Input
        self.in_queue = TaskQueue(self)
        self.in_queue.daemon = True
        
        # Output
        self.out_writer = ResultWriter(self)
        self.out_writer.daemon = True
        
        # Run
        self.runner = TaskRunner(self)
        self.runner.daemon = True
        if self.state == 'started':
            self.unpause()
    
    def quit(self):
        self.pause()
        self.exit = True
        self.resize(0)
        self.pool.close()
    
    def kill(self):
        self.pause()
        self.exit = True
        self.resize(0)
        self.pool.terminate()
    
    def set_shared_size(self, size):
        self.shared_size = size
    
    def pause(self):
        self.state = 'stopped'
    
    def unpause(self):
        self.start()
        self.state = 'started'
    
    @property
    def stopped(self):
        return self.state == 'stopped'
    
    @property
    def trace(self):
        return self.pool.introspect()
    
    def resize(self, size):
        self.pool.resize(size)
        self.num_processes = size
    
    def start(self):
        if self.initialized:
            return
        with print_exc:
            self.out_writer.clear_old()
        if not self.in_queue.is_alive():
            self.in_queue.start()
        if not self.runner.is_alive():
            self.runner.start()
        if not self.out_writer.is_alive():
            self.out_writer.start()
        self.initialized = True

    def json_info(self, detailed=False):
        info = {
            'entry' : self.entry,
            'queue' : self.queue,
            'queue_full' : self.queue_name,
            'state' : self.state,
            'processes' : self.num_processes,
        }
        
        success = self.redis.get(self.success_count)
        if success is None:
            success = 0
        info.update(** {
            'queued' : self.redis.llen(self.queue_name),
            'active' : self.redis.hlen(self.active_queue)+self.redis.llen(self.pending_queue),
            'failed' : self.redis.llen(self.failed_queue),
            'success' : success,
        })
        return info

    
    @classmethod
    def from_conf(self, broker_host, broker_port, queue, data, base_path, paths, environment, instance_id, shared_size, version):
        c_args = {
            'entry' : data.get('entry', queue),
            'queue' : queue,
            'python_paths' : paths,
            'base_path' : base_path,
            'environment' : environment,
            'instance_id' : instance_id,
            'shared_size' : shared_size,
            'version' : version,
        }
        if 'state' in data:
            c_args['state'] = data['state']
        
        if 'processes' in data:
            c_args['num_processes'] = data['processes']
        return Command(broker_host, broker_port, **c_args)
    
    @property
    def trace(self):
        return self.pool.introspect()
    
    def poll(self):
        with self.lock:
            return len(self.pending) == 0 and self.exit
    
    @property
    def failed_items(self):
        import json
        items = []
        for x in self.redis.lrange(self.failed_queue, 0, -1):
            items.append(json.loads(x[5:]))
        return items
    
    @property
    def failed_with_raw(self):
        import json
        for x in self.redis.lrange(self.failed_queue, 0, -1):
            yield (json.loads(x[5:]), x)
    
    def requeue_all(self):
        pipe = self.redis.pipeline(transaction=True)
        for item, raw in self.failed_with_raw:
            BaseRequest.send_raw(self.redis, self.queue_name, item['request'])
            pipe.lrem(self.failed_queue, raw)
        pipe.execute()
        
    
    def requeue_failed(self, id):
        for item, raw in self.failed_with_raw:
            if item['id'] != id:
                continue
            pipe = self.redis.pipeline(transaction=True)
            BaseRequest.send_raw(self.redis, self.queue_name, item['request'])
            pipe.lrem(self.failed_queue, raw)
            pipe.execute()
    
    def remove_all_failed(self):
        self.redis.delete(self.failed_queue)
    
    def remove_failed(self, id):
        for item, raw in self.failed_with_raw:
            if item['id'] != id:
                continue
            self.redis.lrem(self.failed_queue, raw)

class TaskQueue(Thread):
    def __init__(self, command):
        Thread.__init__(self)
        self.command = command
        self.redis = Redis(command.broker_host, port=command.broker_port)
        self.queue = Queue()
    
    def get(self, timeout=None):
        try:
            return self.queue.get(block=True, timeout=timeout)
        except Empty:
            return None
    
    def qsize(self):
        return self.queue.qsize()
    
    def run(self):
        while True:
            if self.command.exit and self.queue.qsize() == 0:
                break
            if self.command.stopped:
                time.sleep(0.5)
                continue
            if self.queue.qsize() >= self.command.num_processes + self.command.shared_size:
                time.sleep(0.05)
                continue
            self.enqueue()
    
    def enqueue(self):
        item = self.redis.brpoplpush(self.command.queue_name, self.command.pending_queue, 1)
        if item is None:
            return
        pipe = self.redis.pipeline(transaction=True)
        pipe.hset(self.command.active_queue, item, self.command.instance_id)
        pipe.lrem(self.command.pending_queue, item, 1)
        pipe.execute()
        self.queue.put(item)
    


class ResultWriter(Thread):
    daemon = True
    def __init__(self, command):
        Thread.__init__(self)
        self.command = command
        self.lock = RLock()
        self.pending = []
        self.redis = Redis(command.broker_host, port=command.broker_port)
    
    def add_pending(self, result):
        with self.lock:
            self.pending.append(result)
        return result
    
    def full(self, pool='private', max=None):
        if max is None:
            max = self.command.num_processes
        with self.lock:
            count = len([x for x in self.pending if x['pool'] == pool])
        return count > max
    
    def shared_full(self):
        return self.full(pool='shared', max=self.command.shared_size)
    
    def run(self):
        while True:
            with print_exc:
                try:
                    if not self.check_pending():
                        time.sleep(0.1)
                except Exit:
                    break
        # clear out anything remaining before exiting
        self.clear_old()
    
    def check_pending(self):
        with self.lock:
            if self.command.exit and len(self.pending) == 0:
                raise Exit()
            for p in self.pending:
                if not p['async'].ready():
                    continue
                break
            else:
                return False
            self.pending = [x for x in self.pending if x['request'].response_queue != p['request'].response_queue]
        
        try:
            request = p['request']
            ret = p['async'].get()
            response = request.response_class(request.response_queue, 'SUCCESS', ret)
            error = False
        except Exception, e:
            response = request.response_class(request.response_queue, 'ERROR', None)
            response.set_exception(e)
            error = True
        self.respond(request, response, error=error)
        return True
    
    def respond(self, request, response, error=False, retry=True):
        if not request.response_queue:
            return
        try:
            pipe = self.redis.pipeline(transaction=True)
            if not request.offline:
                pipe.lpush(request.response_queue, response.serialize_resp())
                pipe.expire(request.response_queue, request.timeout)
            if error:
                pipe.lpush(self.command.failed_queue, response.serialize_failed_resp(request))
            else:
                pipe.incr(self.command.success_count)
            pipe.hdel(self.command.active_queue, request.raw)
            pipe.execute()
        except KeyboardInterrupt:
            if retry:
                self.respond(request, response, error=error, retry=False)
    
    def clear_old(self):
        def handle_old(request, del_cmd):
            request_parsed = BaseRequest.from_serialized(request)
            response = request_parsed.response_class(request_parsed.response_queue, 'INCOMPLETE', None)
            pipe = self.redis.pipeline(transaction=True)
            pipe.lpush(self.command.failed_queue, response.serialize_failed_resp(request_parsed))
            del_cmd(pipe, request)
            pipe.execute()
        with self.lock:
            for request, v in self.redis.hgetall(self.command.active_queue).iteritems():
                if v != self.command.instance_id:
                    continue
                handle_old(request, lambda pipe, k : pipe.hdel(self.command.active_queue, k))
            # this could /technically/ delete something pending, but the pending 
            # queue is used for a very short period of time and if the system is 
            # up long enough to delete the pending item it should be OK
            for request in self.redis.lrange(self.command.pending_queue, 0, -1):
                handle_old(request, lambda pipe, k : pipe.lrem(self.command.pending_queue, k, 1))


class TaskRunner(Thread):
    def __init__(self, command):
        Thread.__init__(self)
        self.command = command
        self.input = self.command.in_queue
        self.output = self.command.out_writer
    
    def run(self):
        while True:
            if self.command.exit:
                break
            if self.command.stopped or self.command.num_processes == 0:
                time.sleep(0.5)
                continue
            with print_exc:
                if self.ready():
                    self.run_fun()
                else:
                    time.sleep(0.05)
    
    def ready(self):
        stopped = self.command.stopped
        full = self.output.full()
        queue_empty = self.input.qsize() == 0
        return not stopped and not full and not queue_empty
    
    def shared_ready(self):
        stopped = self.command.stopped
        full = self.output.shared_full()
        queue_empty = self.input.qsize() == 0
        return not stopped and not full and not queue_empty
    
    def run_fun(self, pool=None, pool_type='private'):
        if pool is None:
            pool = self.command.pool
        
        # Get item
        request = self.input.get()
        if request is None:
            return
        request = BaseRequest.from_serialized(request)
        
        # run fun
        kwds = {
            'entry' : self.command.entry,
            'args' : request.args,
            'kwds' : request.kwds,
            'format' : request.format,
        }
        # args = [self.application.entry]
        # args.extend([a for a in request.args])
        return self.output.add_pending({
            'async' : pool.apply_async(worker_fun, kwds=kwds),
            'request' : request,
            'pool' : pool_type,
        })

class Exit(Exception):
    pass






from datetime import datetime
import sys
import traceback
import uuid
import cPickle as pickle
import json

class Promise(object):
    def __init__(self, redis, req_queue, queue, loader, timeout):
        self.queue = queue
        self.req_queue = req_queue
        self.loader = loader
        self.redis = redis
        self.timeout = timeout
        
    def poll(self):
        return self.redis.llen(self.queue) == 0
    
    def get(self, timeout=0):
        ''' Get the response.  Automatically raise an exception if there was
            an error, return the data otherwise.  Waits for the timeout length
            of the original request unless the timeout parameter is passed'''
        if self.timeout:
            timeout = self.timeout
        response = self.get_response(timeout=timeout)
        if response is None:
            return response
        if response.status == 'ERROR':
            e = Exception(response.error_value)
            e.traceback = response.traceback
            e.type = response.error_type
            raise e
        return response.data
    
    def get_response(self, timeout=0):
        resp = self.redis.brpop(self.queue, timeout=timeout)
        if not resp:
            return None
        _, value = resp
        return self.loader.from_serialized(value)


class BaseRequest(object):
    def __init__(self, redis, queue, args=tuple(), kwds=dict(), timeout=None, offline=False):
        self.queue = queue
        self.args = args
        self.kwds = kwds
        self.redis = redis
        self.response_queue = uuid.uuid4().hex
        self.timeout = timeout
        self.raw = None
        self.offline = offline
    
    @classmethod
    def send_raw(self, redis, queue, request_data):
        redis.lpush(queue, request_data)
    
    def send(self):
        request_data = self.serialize({
            'response_queue' : self.response_queue,
            'args' : self.serialize(self.args),
            'kwds' : self.serialize(self.kwds),
            'queue' : self.queue,
            'offline' : self.offline,
        })
        request_data = '%s:%s' % (self.format, request_data)
        if self.timeout:
            pipe = self.redis.pipeline()
            pipe.lpush(self.queue, request_data)
            # pipe.expire(self.timeout)
            pipe.execute()
        else:
            self.redis.lpush(self.queue, request_data)
        return Promise(self.redis, self.queue, self.response_queue, self.response_class, self.timeout)
    
    @property
    def response_class(self):
        return FORMAT_TO_RESPONSE_CLASS[self.format]
    
    @classmethod
    def from_serialized(cls, ser):
        format, _, data = ser.partition(':')
        assert format in ['json', 'pickle']
        if format == 'json':
            cls = JSONRequest
        elif format == 'pickle':
            cls = PRequest
        deser = cls.deserialize(data)
        instance = cls(None, deser['queue'], args=deser['args'], kwds=deser['kwds'])
        instance.response_queue = deser['response_queue']
        instance.offline = deser.get('offline', False) # temporarily use get for old format compatability
        instance.raw = ser
        return instance
        

class JSONRequest(BaseRequest):
    ''' A request using json to serialize/deserialize '''
    serialize = classmethod(lambda self, x : json.dumps(x))
    deserialize = classmethod(lambda self, x : json.loads(x))
    format = 'json'

class PRequest(BaseRequest):
    ''' A request using cPickle to serialize/deserialize '''
    serialize = classmethod(lambda self, x : pickle.dumps(x))
    deserialize = classmethod(lambda self, x : pickle.loads(x))
    format = 'pickle'

class BaseResponse(object):
    def __init__(self, queue, status, data):
        self.status = status
        self.data = data
        self.queue = queue
        self.error_type = None
        self.traceback = None
        self.error_value = None
    
    def serialize_resp(self):
        response_data = self.serialize(self.response_dict)
        return '%s:%s' % (self.format, response_data)
    
    @property
    def response_dict(self):
        response_data = {
            'status' : self.status,
            'data' : self.data,
            'queue' : self.queue,
            
        }
        if self.status == 'ERROR':
            response_data.update(traceback=self.traceback, error_type=self.error_type, error_value=self.error_value)
        return response_data
    
    def serialize_failed_resp(self, request):
        resp = {
            'id' : uuid.uuid4().hex,
            'response' : self.response_dict,
            'request' : request.raw,
            'time' : datetime.utcnow().isoformat()
        }
        response_data = json.dumps(resp)
        return '%s:%s' % ('json', response_data)

    
    def set_exception(self, e):
        self.traceback = traceback.format_exc()
        if hasattr(e, 'traceback'):
            self.traceback = e.traceback
        self.error_type = type(e).__name__
        self.error_value = repr(e)
    
    @classmethod
    def from_serialized(cls, ser):
        format, _, data = ser.partition(':')
        assert format in ['json', 'pickle']
        if format == 'json':
            cls = JSONResponse
        elif format == 'pickle':
            cls = PResponse
        deser = cls.deserialize(data)
        instance = cls(deser['queue'], deser['status'], deser['data'])
        if instance.status == 'ERROR':
            instance.traceback = deser['traceback']
            instance.error_type = deser['error_type']
            instance.error_value = deser['error_value']
        return instance

class JSONResponse(BaseResponse):
    ''' A request using json to serialize/deserialize '''
    serialize = classmethod(lambda cls, x : json.dumps(x))
    deserialize = classmethod(lambda cls, x : json.loads(x))
    format = 'json'

    def set_exception(self, e):
        self.traceback = traceback.format_exc()
        self.error_type = str(type(e))
        self.error_value = str(e)


class PResponse(BaseResponse):
    ''' A request using cPickle to serialize/deserialize '''
    serialize = classmethod(lambda cls, x : pickle.dumps(x))
    deserialize = classmethod(lambda cls, x : pickle.loads(x))
    format = 'pickle'

FORMAT_TO_RESPONSE_CLASS = {
    'json' : JSONResponse, 
    'pickle' : PResponse
} 